@@ -13,6 +13,8 @@ if ENV['COVERAGE']
 
   ENV.delete('COVERAGE')
   SimpleCov.instance_eval do
+    enable_coverage :branch
+
     start do
       add_filter "/test/"
       add_group('Missing'){|src| src.covered_percent < 100}
@@ -23,8 +25,10 @@ end
 
 require 'erubi'
 require 'erubi/capture_end'
-require 'minitest/spec'
-require 'minitest/autorun'
+
+ENV['MT_NO_PLUGINS'] = '1' # Work around stupid autoloading of plugins
+gem 'minitest'
+require 'minitest/global_expectations/autorun'
 
 describe Erubi::Engine do
   before do
@@ -35,7 +39,8 @@ describe Erubi::Engine do
     t = (@options[:engine] || Erubi::Engine).new(input, @options)
     tsrc = t.src
     eval(tsrc, block.binding).must_equal result
-    tsrc = tsrc.gsub("'.freeze;", "';") if RUBY_VERSION >= '2.1'
+    strip_freeze = defined?(@strip_freeze) ? @strip_freeze : RUBY_VERSION >= '2.1'
+    tsrc = tsrc.gsub(/\.freeze/, '') if strip_freeze
     tsrc.must_equal src
   end
 
@@ -76,8 +81,20 @@ describe Erubi::Engine do
     end
   end
 
+  it "should handle no tags with frozen source" do
+    check_output(<<END1.freeze, <<END2, <<END3){}
+a
+END1
+_buf = ::String.new; _buf << 'a
+';
+_buf.to_s
+END2
+a
+END3
+  end
+
   it "should handle no options" do
-    list = ['&\'<>"2']
+    list = list = ['&\'<>"2']
     check_output(<<END1, <<END2, <<END3){}
 <table>
  <tbody>
@@ -92,7 +109,7 @@ describe Erubi::Engine do
 </table>
 <%== i+1 %>
 END1
-_buf = String.new; _buf << '<table>
+_buf = ::String.new; _buf << '<table>
  <tbody>
 ';   i = 0
      list.each_with_index do |item, i| 
@@ -119,43 +136,89 @@ END2
 END3
   end
 
+  it "should escape all backslashes and apostrophes in text" do
+    list = list = ['&\'<>"2']
+    check_output(<<END1.chomp, <<END2, <<END3){}
+<table>
+ <tbody>' ' \\ \\
+  <% i = 0
+     list.each_with_index do |item, i| %>
+  <tr>
+   <td><%= i+1 -%>
+</td>
+   <td><%== item %></td>
+  </tr>
+ <% end %>
+ </tbody>
+</table>
+<%== i+1 %>
+<%
+%>
+END1
+_buf = ::String.new; _buf << '<table>
+ <tbody>\\' \\' \\\\ \\\\
+';   i = 0
+     list.each_with_index do |item, i| 
+ _buf << '  <tr>
+   <td>'; _buf << ( i+1 ).to_s; _buf << '</td>
+   <td>'; _buf << ::Erubi.h(( item )); _buf << '</td>
+  </tr>
+';  end 
+ _buf << ' </tbody>
+</table>
+'; _buf << ::Erubi.h(( i+1 )); _buf << '
+';
+_buf.to_s
+END2
+<table>
+ <tbody>' ' \\ \\
+  <tr>
+   <td>1</td>
+   <td>&amp;&#39;&lt;&gt;&quot;2</td>
+  </tr>
+ </tbody>
+</table>
+1
+END3
+  end
+
   it "should strip only whitespace for <%, <%- and <%# tags" do
     check_output(<<END1, <<END2, <<END3){}
-  <% 1 %>  
+  <% a = 1 %>  
 a
-  <%- 2 %>  
+  <%- a = 2 %>  
 b
-  <%# 3 %>  
+  <%# a = 3 %>  
 c
- /<% 1 %>  
+ /<% a = 1 %>  
 a
-/ <%- 2 %>  
+/ <%- a = 2 %>  
 b
-//<%# 3 %>  
+//<%# a = 3 %>  
 c
-  <% 1 %> /
+  <% a = 1 %> /
 a
-  <%- 2 %>/ 
+  <%- a = 2 %>/ 
 b
-  <%# 3 %>//
+  <%# a = 3 %>//
 c
 END1
-_buf = String.new;   1   
+_buf = ::String.new;   a = 1   
  _buf << 'a
-';   2   
+';   a = 2   
  _buf << 'b
 ';
  _buf << 'c
- /'; 1 ; _buf << '  
+ /'; a = 1 ; _buf << '  
 '; _buf << 'a
-/ '; 2 ; _buf << '  
+/ '; a = 2 ; _buf << '  
 '; _buf << 'b
 //';
  _buf << '  
 '; _buf << 'c
-'; _buf << '  '; 1 ; _buf << ' /
+'; _buf << '  '; a = 1 ; _buf << ' /
 a
-'; _buf << '  '; 2 ; _buf << '/ 
+'; _buf << '  '; a = 2 ; _buf << '/ 
 b
 '; _buf << '  ';; _buf << '//
 c
@@ -181,7 +244,7 @@ END3
   end
 
   it "should handle ensure option" do
-    list = ['&\'<>"2']
+    list = list = ['&\'<>"2']
     @options[:ensure] = true
     @options[:bufvar] = '@a'
     @a = 'bar'
@@ -199,7 +262,7 @@ END3
 </table>
 <%== i+1 %>
 END1
-begin; __original_outvar = @a if defined?(@a); @a = String.new; @a << '<table>
+begin; __original_outvar = @a#{' if defined?(@a)' if RUBY_VERSION < '3'}; @a = ::String.new; @a << '<table>
  <tbody>
 ';   i = 0
      list.each_with_index do |item, i| 
@@ -230,6 +293,195 @@ END3
     @a.must_equal 'bar'
   end
 
+  it "should handle chain_appends option" do
+    @options[:chain_appends] = true
+    list = list = ['&\'<>"2']
+    check_output(<<END1, <<END2, <<END3){}
+<table>
+ <tbody>
+  <% i = 0
+     list.each_with_index do |item, i| %>
+  <tr>
+   <td><%= i+1 %></td>
+   <td><%== item %></td>
+  </tr>
+ <% end %>
+ </tbody>
+</table>
+<%== i+1 %>
+END1
+_buf = ::String.new;; _buf << '<table>
+ <tbody>
+';    i = 0
+     list.each_with_index do |item, i| 
+; _buf << '  <tr>
+   <td>' << ( i+1 ).to_s << '</td>
+   <td>' << ::Erubi.h(( item )) << '</td>
+  </tr>
+';   end 
+; _buf << ' </tbody>
+</table>
+' << ::Erubi.h(( i+1 )) << '
+'
+; _buf.to_s
+END2
+<table>
+ <tbody>
+  <tr>
+   <td>1</td>
+   <td>&amp;&#39;&lt;&gt;&quot;2</td>
+  </tr>
+ </tbody>
+</table>
+1
+END3
+  end
+
+  it "should handle :freeze_template_literals => true option" do
+    @options[:freeze_template_literals] = true
+    list = list = ['&\'<>"2']
+    @strip_freeze = false
+    check_output(<<END1, <<END2, <<END3){}
+<table>
+ <tbody>
+  <% i = 0
+     list.each_with_index do |item, i| %>
+  <tr>
+   <td><%= i+1 %></td>
+   <td><%== item %></td>
+  </tr>
+ <% end %>
+ </tbody>
+</table>
+<%== i+1 %>
+END1
+_buf = ::String.new; _buf << '<table>
+ <tbody>
+'.freeze;   i = 0
+     list.each_with_index do |item, i| 
+ _buf << '  <tr>
+   <td>'.freeze; _buf << ( i+1 ).to_s; _buf << '</td>
+   <td>'.freeze; _buf << ::Erubi.h(( item )); _buf << '</td>
+  </tr>
+'.freeze;  end 
+ _buf << ' </tbody>
+</table>
+'.freeze; _buf << ::Erubi.h(( i+1 )); _buf << '
+'.freeze;
+_buf.to_s
+END2
+<table>
+ <tbody>
+  <tr>
+   <td>1</td>
+   <td>&amp;&#39;&lt;&gt;&quot;2</td>
+  </tr>
+ </tbody>
+</table>
+1
+END3
+  end
+
+  it "should handle :freeze_template_literals => false option" do
+    @options[:freeze_template_literals] = false
+    list = list = ['&\'<>"2']
+    @strip_freeze = false
+    check_output(<<END1, <<END2, <<END3){}
+<table>
+ <tbody>
+  <% i = 0
+     list.each_with_index do |item, i| %>
+  <tr>
+   <td><%= i+1 %></td>
+   <td><%== item %></td>
+  </tr>
+ <% end %>
+ </tbody>
+</table>
+<%== i+1 %>
+END1
+_buf = ::String.new; _buf << '<table>
+ <tbody>
+';   i = 0
+     list.each_with_index do |item, i| 
+ _buf << '  <tr>
+   <td>'; _buf << ( i+1 ).to_s; _buf << '</td>
+   <td>'; _buf << ::Erubi.h(( item )); _buf << '</td>
+  </tr>
+';  end 
+ _buf << ' </tbody>
+</table>
+'; _buf << ::Erubi.h(( i+1 )); _buf << '
+';
+_buf.to_s
+END2
+<table>
+ <tbody>
+  <tr>
+   <td>1</td>
+   <td>&amp;&#39;&lt;&gt;&quot;2</td>
+  </tr>
+ </tbody>
+</table>
+1
+END3
+  end
+
+  it "should handle ensure option with no bufvar" do
+    list = list = ['&\'<>"2']
+    @options[:ensure] = true
+    check_output(<<END1, <<END2, <<END3){}
+<table>
+ <tbody>
+  <% i = 0
+     list.each_with_index do |item, i| %>
+  <tr>
+   <td><%= i+1 %></td>
+   <td><%== item %></td>
+  </tr>
+ <% end %>
+ </tbody>
+</table>
+<%== i+1 %>
+END1
+begin; __original_outvar = _buf if defined?(_buf); _buf = ::String.new; _buf << '<table>
+ <tbody>
+';   i = 0
+     list.each_with_index do |item, i| 
+ _buf << '  <tr>
+   <td>'; _buf << ( i+1 ).to_s; _buf << '</td>
+   <td>'; _buf << ::Erubi.h(( item )); _buf << '</td>
+  </tr>
+';  end 
+ _buf << ' </tbody>
+</table>
+'; _buf << ::Erubi.h(( i+1 )); _buf << '
+';
+_buf.to_s
+; ensure
+  _buf = __original_outvar
+end
+END2
+<table>
+ <tbody>
+  <tr>
+   <td>1</td>
+   <td>&amp;&#39;&lt;&gt;&quot;2</td>
+  </tr>
+ </tbody>
+</table>
+1
+END3
+  end
+
+  it "should handle trailing rspace with - modifier in <%|= and <%|" do
+    eval(::Erubi::CaptureEndEngine.new("<%|= '&' -%>\n<%| -%>\n").src).must_equal '&'
+  end
+
+  it "should handle lspace in <%|=" do
+    eval(::Erubi::CaptureEndEngine.new("<%|= %><%| %><%|= %><%| %>").src).must_equal ''
+  end
+
   it "should have <%|= with CaptureEndEngine not escape by default" do
     eval(::Erubi::CaptureEndEngine.new('<%|= "&" %><%| %>').src).must_equal '&'
     eval(::Erubi::CaptureEndEngine.new('<%|= "&" %><%| %>', :escape=>false).src).must_equal '&'
@@ -263,9 +515,9 @@ END3
  </tbody>
 </table>
 END1
-#{'__erubi = ::Erubi;' unless escape}@a = String.new; @a << '<table>
+#{'__erubi = ::Erubi; ' unless escape}@a = ::String.new; @a << '<table>
  <tbody>
-'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = String.new; __erubi_stack.last << (( bar do  @a << '
+'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << (( bar do  @a << '
 '; @a << '   <b>'; @a << #{!escape ? '__erubi' : '::Erubi'}.h(( '&' )); @a << '</b>
 '; @a << ' '; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
 '; @a << ' </tbody>
@@ -299,9 +551,9 @@ END3
  </tbody>
 </table>
 END1
-#{'__erubi = ::Erubi;' if escape}@a = String.new; @a << '<table>
+#{'__erubi = ::Erubi; ' if escape}@a = ::String.new; @a << '<table>
  <tbody>
-'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( bar do  @a << '
+'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( bar do  @a << '
 '; @a << '   <b>'; @a << #{escape ? '__erubi' : '::Erubi'}.h(( '&' )); @a << '</b>
 '; @a << ' '; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
 '; @a << ' </tbody>
@@ -333,9 +585,9 @@ END3
  </tbody>
 </table>
 END1
-#{'__erubi = ::Erubi;' if escape}@a = String.new; @a << '<table>
+#{'__erubi = ::Erubi; ' if escape}@a = ::String.new; @a << '<table>
  <tbody>
-'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( quux do |i|  @a << '
+'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( quux do |i|  @a << '
 '; @a << '   <b>'; @a << #{escape ? '__erubi' : '::Erubi'}.h(( "\#{i}&" )); @a << '</b>
 '; @a << ' '; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
 '; @a << ' </tbody>
@@ -372,11 +624,11 @@ END3
  </tbody>
 </table>
 END1
-#{'__erubi = ::Erubi;' if escape}@a = String.new; @a << '<table>
+#{'__erubi = ::Erubi; ' if escape}@a = ::String.new; @a << '<table>
  <tbody>
-'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( bar do  @a << '
+'; @a << '  ';begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( bar do  @a << '
 '; @a << '   <b>'; @a << #{escape ? '__erubi' : '::Erubi'}.h(( '&' )); @a << '</b>
-'; @a << '   ';begin; (__erubi_stack ||= []) << @a; @a = String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( baz do  @a << 'e'; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
+'; @a << '   ';begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << #{escape ? '__erubi' : '::Erubi'}.h(( baz do  @a << 'e'; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
 '; @a << ' '; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
 '; @a << ' </tbody>
 </table>
@@ -400,7 +652,7 @@ END3
       @options[var] = "@_out_buf"
       @options[:freeze] = true
       @items = [2]
-      i = 0
+      i = i = 0
       check_output(<<END1, <<END2, <<END3){}
 <table>
   <% for item in @items %>
@@ -412,7 +664,7 @@ END3
 </table>
 END1
 # frozen_string_literal: true
-@_out_buf = String.new; @_out_buf << '<table>
+@_out_buf = ::String.new; @_out_buf << '<table>
 ';   for item in @items 
  @_out_buf << '  <tr>
     <td>'; @_out_buf << ( i+1 ).to_s; @_out_buf << '</td>
@@ -435,7 +687,7 @@ END3
 
   it "should handle <%% and <%# syntax" do
     @items = [2]
-    i = 0
+    i = i = 0
     check_output(<<END1, <<END2, <<END3){}
 <table>
 <%% for item in @items %>
@@ -446,7 +698,7 @@ END3
   <%% end %>
 </table>
 END1
-_buf = String.new; _buf << '<table>
+_buf = ::String.new; _buf << '<table>
 '; _buf << '<% for item in @items %>
 '; _buf << '  <tr>
     <td>';; _buf << '</td>
@@ -468,10 +720,44 @@ END2
 END3
   end
 
+  it "should handle <%% with a different literal prefix/postfix" do
+    @options[:literal_prefix] = "{%"
+    @options[:literal_postfix] = "%}"
+    @items = [2]
+    i = i = 0
+    check_output(<<END1, <<END2, <<END3){}
+<table>
+  <%% for item in @items %>
+  <tr>
+  </tr>
+  <%% end %>
+  <%%= "literal" %>
+</table>
+END1
+_buf = ::String.new; _buf << '<table>
+'; _buf << '  {% for item in @items %}
+'; _buf << '  <tr>
+  </tr>
+'; _buf << '  {% end %}
+'; _buf << '  {%= "literal" %}
+'; _buf << '</table>
+';
+_buf.to_s
+END2
+<table>
+  {% for item in @items %}
+  <tr>
+  </tr>
+  {% end %}
+  {%= "literal" %}
+</table>
+END3
+  end
+
   it "should handle :trim => false option" do
     @options[:trim] = false
     @items = [2]
-    i = 0
+    i = i = 0
     check_output(<<END1, <<END2, <<END3){}
 <table>
   <% for item in @items %>
@@ -482,11 +768,11 @@ END3
     <td><%== item %></td>
   </tr>
   <% end %><%#%>
-  <% i %>a
-  <% i %>
+  <% i = 1 %>a
+  <% i = 1 %>
 </table>
 END1
-_buf = String.new; _buf << '<table>
+_buf = ::String.new; _buf << '<table>
 '; _buf << '  '; for item in @items ; _buf << '
 '; _buf << '  <tr>
     <td>';
@@ -496,8 +782,8 @@ _buf = String.new; _buf << '<table>
   </tr>
 '; _buf << '  '; end ;
  _buf << '
-'; _buf << '  '; i ; _buf << 'a
-'; _buf << '  '; i ; _buf << '
+'; _buf << '  '; i = 1 ; _buf << 'a
+'; _buf << '  '; i = 1 ; _buf << '
 '; _buf << '</table>
 ';
 _buf.to_s
@@ -519,8 +805,8 @@ END3
     it "should handle :#{opt} and :escapefunc options" do
       @options[opt] = true
       @options[:escapefunc] = 'h.call'
-      h = proc{|s| s.to_s*2}
-      list = ['2']
+      h = h = proc{|s| s.to_s*2}
+      list = list = ['2']
       check_output(<<END1, <<END2, <<END3){}
 <table>
  <tbody>
@@ -535,7 +821,7 @@ END3
 </table>
 <%== i+1 %>
 END1
-_buf = String.new; _buf << '<table>
+_buf = ::String.new; _buf << '<table>
  <tbody>
 ';   i = 0
      list.each_with_index do |item, i| 
@@ -565,7 +851,7 @@ END3
 
   it "should handle :escape option without :escapefunc option" do
     @options[:escape] = true
-    list = ['&\'<>"2']
+    list = list = ['&\'<>"2']
     check_output(<<END1, <<END2, <<END3){}
 <table>
  <tbody>
@@ -579,7 +865,7 @@ END3
  </tbody>
 </table>
 END1
-__erubi = ::Erubi;_buf = String.new; _buf << '<table>
+__erubi = ::Erubi; _buf = ::String.new; _buf << '<table>
  <tbody>
 ';   i = 0
      list.each_with_index do |item, i| 
@@ -607,7 +893,7 @@ END3
   it "should handle :preamble and :postamble options" do
     @options[:preamble] = '_buf = String.new("1");'
     @options[:postamble] = "_buf[0...18]\n"
-    list = ['2']
+    list = list = ['2']
     check_output(<<END1, <<END2, <<END3){}
 <table>
  <tbody>
@@ -651,6 +937,12 @@ END3
     Erubi::Engine.new('', :outvar=>'foo').bufvar.must_equal 'foo'
   end
 
+  it "should work with BasicObject methods" do
+    c = Class.new(BasicObject)
+    c.class_eval("def a; #{Erubi::Engine.new('2').src} end")
+    c.new.a.must_equal '2'
+  end if defined?(BasicObject)
+
   it "should return frozen object" do
     Erubi::Engine.new('').frozen?.must_equal true
   end
@@ -663,4 +955,110 @@ END3
     proc{Erubi::Engine.new('<%] %>', :regexp =>/<%(={1,2}|\]|-|\#|%)?(.*?)([-=])?%>([ \t]*\r?\n)?/m)}.must_raise ArgumentError
     proc{Erubi::CaptureEndEngine.new('<%] %>', :regexp =>/<%(={1,2}|\]|-|\#|%)?(.*?)([-=])?%>([ \t]*\r?\n)?/m)}.must_raise ArgumentError
   end
+
+  it "should respect the :yield_returns_buffer option for making templates return the (potentially modified) buffer" do
+    @options[:engine] = ::Erubi::CaptureEndEngine
+    @options[:bufvar] = '@a'
+
+    def self.bar
+      a = String.new
+      a << "a"
+      yield 'burgers'
+      case b = (yield 'salads')
+      when String
+        a << b
+        a << 'b'
+        a.upcase
+      end
+    end
+
+    check_output(<<END1, <<END2, <<END3){}
+<%|= bar do |item| %>
+Let's eat <%= item %>!
+<% nil %><%| end %>
+END1
+@a = ::String.new;begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << (( bar do |item|  @a << '
+'; @a << 'Let\\'s eat '; @a << ( item ).to_s; @a << '!
+'; nil ; end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
+';
+@a.to_s
+END2
+
+END3
+
+    @options[:yield_returns_buffer] = true
+
+    check_output(<<END1, <<END2, <<END3) {}
+<%|= bar do |item| %>
+Let's eat <%= item %>!
+<% i = i = nil %><%| end %>
+END1
+@a = ::String.new;begin; (__erubi_stack ||= []) << @a; @a = ::String.new; __erubi_stack.last << (( bar do |item|  @a << '
+'; @a << 'Let\\'s eat '; @a << ( item ).to_s; @a << '!
+'; i = i = nil ; @a;  end )).to_s; ensure; @a = __erubi_stack.pop; end; @a << '
+';
+@a.to_s
+END2
+A
+LET'S EAT BURGERS!
+
+LET'S EAT SALADS!
+B
+END3
+  end
+
+  it "should respect the :yield_returns_buffer option for making templates return the (potentially modified) buffer as the result of the block" do
+    @options[:engine] = ::Erubi::CaptureEndEngine
+    @options[:yield_returns_buffer] = true
+
+    def self.bar(foo = nil)
+      if foo.nil?
+        yield
+      else
+        foo
+      end
+    end
+
+    check_output(<<END1, <<END2, <<END3) {}
+<%|= bar do %>
+Let's eat the tacos!
+<%| end %>
+
+Delicious!
+END1
+_buf = ::String.new;begin; (__erubi_stack ||= []) << _buf; _buf = ::String.new; __erubi_stack.last << (( bar do  _buf << '
+'; _buf << 'Let\\'s eat the tacos!
+'; _buf;  end )).to_s; ensure; _buf = __erubi_stack.pop; end; _buf << '
+'; _buf << '
+Delicious!
+';
+_buf.to_s
+END2
+
+Let's eat the tacos!
+
+
+Delicious!
+END3
+
+    check_output(<<END1, <<END2, <<END3) {}
+<%|= bar("Don't eat the burgers!") do %>
+Let's eat burgers!
+<%| end %>
+
+Delicious!
+END1
+_buf = ::String.new;begin; (__erubi_stack ||= []) << _buf; _buf = ::String.new; __erubi_stack.last << (( bar(\"Don't eat the burgers!\") do  _buf << '
+'; _buf << 'Let\\'s eat burgers!
+'; _buf;  end )).to_s; ensure; _buf = __erubi_stack.pop; end; _buf << '
+'; _buf << '
+Delicious!
+';
+_buf.to_s
+END2
+Don't eat the burgers!
+
+Delicious!
+END3
+  end
 end
